
; AT Keyboard host (i8042) emulation v1.0  * preliminary *
; v1.0	96-08-13 (C) ARNE Computers	All rights reserved
; written by Mauricio Culibrk
;
; contact: mauricio@arne.si	 www.arne.si/~mauricio


; this routine is using 4 pins (2 inputs & 2 outputs) to control
; keyboard's 2 bidirectional OC lines (CLK & DATA). The following
; 'drawing' conceptually shows how to connect the related pins/lines
;
;			     vcc	vcc
;			      | 	 |
;			      \ 	-+-
;			      / 2K2	/_\  1N4148
;			      \ 	 |
; hoCLKi    ------------------o---o------o---------   kbd CLOCK
;			      |   |	 |
;			 2N2222   |50pF -+-
; hoCLKo    ______/\/\/\____|/	 ===	/_\
;		    2K2     |\>   |	 |
;			      |   |	 |
;			     /// ///	///
;
; an identical circuit is used for the DATA line. This diagram was
; 'copied' from an classic AT keyboard. I did not use the clamping
; diodes because they are already included as standard pin protection
; in PC microcontrollers. This can be simplified more if the
; related pin is switched between input & output low (simulate OC)
;
; if you need more information about PC AT/XT keyboard specification,
; take a look at www.paranoia.com/~filipg
;
; NOTE: if you implement/use this routine, please let me know if you
; encunter any incompatibility or any other problems. I'll be happy to
; expand/share my experience about this matter. Thanx.



; constants for clock rate of 9600 BPS

#define  hoBit0TO   255 	       ; first bit timeout   (26.5ms)
#define  hoBitXTO   5		       ; other bits timeout  (520uS)
#define  hoPreTX    144 ;; 15 (1.5 ms) OK ; Pre TX delay	(15.0ms)

#define  hoCLKi   PORTB, 0	       ; line Clock as input
#define  hoCLKo   PORTB, 1	       ; line Clock as output
#define  hoDATAi  PORTB, 2	       ; line Data as input
#define  hoDATAo  PORTB, 3	       ; line Data as output

#define  ODDPARITY		       ; use odd parity

#define  hoParity hoCount, 7	       ; calculated parity bit

#define  hoTimerON  hoFlags, 0	       ; IO Engine Timer active flag
#define  hoSchedON  hoFlags, 1	       ; IO Engine Scheduler active
#define  hoRWMode   hoFlags, 2	       ; IO Engine mode, RX=0, TX=1
#define  hoError    hoFlags, 3	       ; PErr, TOut, Abort Error flag
#define  hoBusy     hoFlags, 4	       ; IO Engine Busy
#define  hoBufFul   hoFlags, 5	       ; byte rcvd, rdy to xmit

 New hoFlags			       ; bitmap (state,timer,error...)
 New hoCount			       ; bit counter & parity generator
 New hoBuffer			       ; data buffer
 New hoTmrCount 		       ; io timer counter


#include "p16c84.inc"
#include "custom.h"

; pin handling macros
#include "ioPins.inc"




; *****Init procedure

HostInit:
	      SetRPage 1
	      bcf    OPTION_REG, INTEDG  ; set interrupt on falling edge
	      bsf    hoCLKi		 ; set inputs
	      bsf    hoDATAi
	      bcf    hoCLKo		 ; set outputs
	      bcf    hoDATAo
	      SetRPage 0
	      clrf   hoFlags		 ; preset runMode
	      jmp    HOSTsetBusy	 ; on power-on enter busy state


; ***** Timer Engine
; called by TMR interrupt

HostTimer:
	      btfss  hoTimerON		    ; check if timer functions active
	      ret
	      decfsz hoTmrCount 	    ; wait for event
	      ret
	      btfss  hoSchedON		    ; is this a Scheduler event?
	      jmp    HstTime		    ; no
	      bcf    hoSchedON		    ; stop scheduler
	      jmp    HOSTsetTX		    ; PreTX -> TX State
HstTime:
	      ;; do something if error occoured before TOut?
	      bcf    hoTimerON		  ; stop timer
	      bcf    hoBusy		  ; error -> done with anything
	      bsf    hoError		  ; set TOut error
	      btfss  hoRWMode		  ; check if RX or TX
	      jmp    HOSTsetBusy	  ; was TX err, stop TX, exit
	      ;; restart TX? for max N times?

	      movfw  hoCount		  ; was the 10th bit rcvd?
	      andlw  H'0F'
	      sublw  10
	      skpc			  ; if cf=1 ioCount >=10, byte rcvd
	      jmp    HOSTsetRX		  ; if cf=0, discard, restart RX
	      bsf    hoBufFul		  ; signal that rx buff full
	      jmp    HOSTsetBusy	  ; terminate RX




; ***** Host IO  Engine code
; called by INT interrupt

HostEngine:   movlw  hoBitXTO		  ; reset TOut timer
	      movwf  hoTmrCount
	      movfw  hoCount		  ; get current bit
	      andlw  H'0F'                ; mask out parity
	      btfsc  hoRWMode		  ; if TX(1), move to next table
	      addlw  11 		  ; TX Table Offset
	      SwitchW
	      jmp    HostRXBit0 	  ; RX Start Bit
	      jmp    HostRXBitX 	  ; RX Data bit 0
	      jmp    HostRXBitX
	      jmp    HostRXBitX
	      jmp    HostRXBitX
	      jmp    HostRXBitX
	      jmp    HostRXBitX
	      jmp    HostRXBitX
	      jmp    HostRXBitX 	  ; RX Data bit 7
	      jmp    HostRXBitP 	  ; RX parity bit
	      jmp    HostRXBitS 	  ; RX stop bit

	      jmp    HostTXBit0 	  ; TX Data bit 0
	      jmp    HostTXBitX
	      jmp    HostTXBitX
	      jmp    HostTXBitX
	      jmp    HostTXBitX
	      jmp    HostTXBitX
	      jmp    HostTXBitX
	      jmp    HostTXBitX 	  ; TX Data bit 7
	      jmp    HostTXBitP 	  ; TX Parity
	      jmp    HostTXBitS 	  ; TX Stop bit
	      jmp    HostTXBitA 	  ; TX Ack bit



; ***** Receive procedures for bits 0..11
; Start bit
HostRXBit0:   bsf    hoTimerON		    ; activate IO Time Counter
	      bcf    hoSchedON
	      bsf    hoBusy		    ; signalize busy to others
#ifdef ODDPARITY
	      bsf    hoParity		    ; generate ODD parity
#else
	      bcf    hoParity		    ; generate EVEN parity
#endif
	      incf   hoCount		    ; prepare for next bit
	      btfss  hoDATAi		    ; test for start bit (lo)
	      ret			    ; exit if OK
	      bcf    hoTimerON		    ; bad start bit, discard, try again
	      jmp    HOSTsetRX		    ; reset to phase 0 of RX state

; Data Bits
HostRXBitX:   setc			    ; assume data=1, CF=1
	      btfss  hoDATAi		    ; if data=0, CF=0
	      clrc
	      rrf    hoBuffer		    ; store bit in MSB, shift
	      movfw  hoBuffer		    ; copy last bit to CF
	      andlw  H'80'                  ; mask last bit
	      xorwf  hoCount		    ; calculate parity
	      incf   hoCount		    ; next bit
	      ret
; Parity Bit
HostRXBitP:   setc			    ; set CF=DATA
	      btfss  hoDATAi
	      clrc
	      rrfw   hoBuffer		    ; make bit 7 of W = DATA
	      xorfw  hoCount		    ; check if DATA.P = Calc.P
	      andlw  H'80'                  ; mask bit 7 (PARITY)
	      skpz			    ; if ZF=1, parity OK
	      bsf    hoError
	      incf   hoCount		    ; next bit
	      ret
; Stop Bit
HostRXBitS:   bcf    hoTimerON		    ; stop TOut
	      ;; btfsc	hoError 	     ; on rx error not store in buffer
	      ;; movfw	hoBuffer	     ; store to buffer
	      ;; movwf	RXBuffer
	      bsf    hoBufFul		    ; signal rx buffer full
	      jmp    HOSTsetBusy	    ; Acknowledge received byte



; ***** Transmit procedures for bits 0..10
; Data bit 0
HostTXBit0:
#ifdef ODDPARITY
	      bsf    hoParity		    ; generate ODD parity
#else
	      bcf    hoParity		    ; otherwise, EVEN parity
#endif

HostTXBitX:   rrf    hoBuffer		    ; shift out LSB
	      skpc
	      PinLO  hoDATAo		    ; if CF=0, Bit=0
	      skpnc
	      PinHI  hoDATAo		    ; if CF=1, Bit=1
	      rrfw   hoBuffer		    ; W's bit 7 <= CF
	      andlw  H'80'
	      xorwf  hoCount		    ; calculate parity
	      incf   hoCount		    ; next bit
	      ret
; Parity Bit
HostTXBitP:   rlfw   hoCount		    ; get Parity in CF
	      skpc
	      PinLO  hoDATAo		    ; output parity=0
	      skpnc
	      PinHI  hoDATAo		    ; output parity=1
	      incf   hoCount		    ; next bit
	      ret
; Stop bit
HostTXBitS:   PinHI  hoDATAo		    ; set Data=High, stop
	      incf   hoCount		    ; next bit
	      ;; check if DATA HI?
	      ret
; Ack bit
HostTXBitA:   bcf    hoTimerON		    ; stop timeout timer
	      bcf    hoBusy		    ; done with TX
	      bcf    hoBufFul		    ; signal tx buffer empty
	      jmp    HOSTsetBusy	    ; terminate TX sequence



; ********** Host Mode Switch Procedures

HOSTsetBusy:
;;		bcf    INTCON, INTE	      ; disable int od CLK line
;;		PinLO  hoCLKo		      ; set CLK lo, DATA high -> busy
;;		PinHI  hoDATAo
;;		bcf    INTCON, INTF	      ; clear pending CLK int
;;		bcf    hoBusy		      ; signal ioEngine done
;;		ret

	      bcf    INTCON, GIE
	      bcf    INTCON, INTE	    ; disable int od CLK line
	      PinLO  hoCLKo		      ; set CLK lo, DATA high -> busy
	      PinHI  hoDATAo
	      bcf    INTCON, INTF	    ; clear pending CLK int
	      bcf    hoBusy		    ; signal ioEngine done
	      bsf    INTCON, GIE
	      ret


HOSTsetRX:
	      bcf    hoRWMode		    ; set Io Engine to RX mode
	      bcf    hoError		    ; clear any 'old' errors
	      bcf    hoBufFul		    ; set ioBufer RW
	      clrf   hoCount		    ; clear bit counter
	      PinHI  hoCLKo		    ; set CLK & DATA high
	      PinHI  hoDATAo
	      bcf    INTCON, INTF	    ; clear pending int & enable
	      bsf    INTCON, INTE
	      ret

HOSTsetPreTX:
	      bsf    hoRWMode		    ; set Io Engine to TX mode
	      bcf    hoError		    ; clear any 'old' errors
	      bsf    hoBusy		    ; signal io engine busy
	      bsf    hoBufFul		    ; signalize buffer write protect
	      movlw  hoPreTX		    ; set TX request delay
	      movwf  hoTmrCount
	      bcf    INTCON, INTE	    ; disable int
	      PinLO  hoCLKo		    ; set CLK & DATA lo
	      PinLO  hoDATAo
	      bcf    INTCON, INTF	    ; clear pending int
	      bsf    hoTimerON
	      bsf    hoSchedON		    ; start scheduler
	      ret

HOSTsetTX:
	      movlw  hoBit0TO		    ; set TO for first TX bit
	      movwf  hoTmrCount
	      bsf    hoTimerON		    ; activate timer
	      clrf   hoCount		    ; clear bit & parity generator
	      PinHI  hoCLKo		    ; set CLK hi, DATA lo
	      bcf    INTCON, INTF	    ; clear pending & enable int
	      bsf    INTCON, INTE
	      ret




